use std::str::FromStr;
use std::sync::{Arc, Weak};
use std::time::Duration;

use bytes::Bytes;
use futures_util::SinkExt;
use futures_util::stream::SplitSink;
use gst;
use gst::{StateChangeError, StateChangeSuccess};
use gst::prelude::*;
use tokio::net::TcpStream;
use tokio::sync::Mutex;
use tokio_tungstenite::tungstenite::Message;
use tokio_tungstenite::WebSocketStream;
use webrtc::media::Sample;
use webrtc::peer_connection::RTCPeerConnection;
use webrtc::track::track_local::track_local_static_sample::TrackLocalStaticSample;

use crate::models;
use crate::models::{Session, SinkPad, WsMessage};
use crate::rtc::SESSIONS;

const FORMAT_TIME: &str = "3";
const TEXT_VERTICAL_BOTTOM: &str = "1";
const TEXT_HORIZONTAL_CENTER: &str = "1";
const TEXT_WHITE_COLOR: &str = "0xFFFFFFFF";
const VIDEO_CAPS: &str = "application/x-rtp,media=video,clock-rate=90000,encoding-name=H264,payload=96";
// const AUDIO_CAPS: &str = "application/x-rtp,media=audio,clock-rate=48000,encoding-name=OPUS,payload=111";

pub fn init() {
    gst::init().expect("failed to initial gstreamer");
}

impl Session {
    pub fn start(&self) -> Result<StateChangeSuccess, StateChangeError> {
        self.pipeline.set_state(gst::State::Playing)
    }

    pub fn pause(&self) -> Result<StateChangeSuccess, StateChangeError> {
        self.pipeline.set_state(gst::State::Paused)
    }

    pub fn stop(&self) -> Result<StateChangeSuccess, StateChangeError> {
        self.pipeline.set_state(gst::State::Null)
    }

    pub fn row_video_count(&self, sinks_length: usize) -> i32 {
        (sinks_length as f64).sqrt().ceil() as i32
    }

    pub fn remove_webcam(&self, video_name: &str) -> Result<(), &str> {
        // if let Err(_) = self.pause() {
        //     return Err("FAILED TO PAUSE THE PIPELINE");
        // }

        match self.pipeline.by_name((video_name.to_owned() + "_text").as_str()) {
            None => {
                return Err("FAILED TO FIND THE TEXT OVERLAY ELEMENT");
            }
            Some(textoverlay) => {
                let src_pad = textoverlay.static_pad("src").unwrap();
                let probe = src_pad.add_probe(gst::PadProbeType::BLOCK_DOWNSTREAM, |_pad, _info| {
                    gst::PadProbeReturn::Ok
                }).unwrap();
                textoverlay.unlink(&self.compositor);
                src_pad.remove_probe(probe);

                if let Err(e) = textoverlay.set_state(gst::State::Null) {
                    return Err("FAILED TO SET TEXT OVERLAY STATE TO NULL");
                }
                if let Err(e) = self.pipeline.remove(&textoverlay) {
                    return Err("FAILED TO REMOVE TEXT OVERLAY FROM PIPELINE");
                }

                let element_names = ["_caps", "_scaler", "_decoder", "_depay", "_queue", "_src"];
                for name in element_names {
                    if let Some(el) = self.pipeline.by_name((video_name.to_owned() + name).as_str()) {
                        el.set_state(gst::State::Null).unwrap();
                        self.pipeline.remove(&el).unwrap();
                    }
                }
            }
        };

        // if let Err(_) = self.start() {
        //     return Err("FAILED TO START THE PIPELINE");
        // };

        // self.pipeline.debug_to_dot_file(DebugGraphDetails::ALL, "test");

        Ok(())
    }

    pub fn change_video_size(&self, video_name: Option<&str>, width: i32, height: i32) -> Result<(), &str> {
        let name = video_name.unwrap_or("output");
        return match self.pipeline.by_name(&format!("{}_caps", name)) {
            None => {
                Err("FAILED TO FIND THE CAPS FILTER ELEMENT")
            }
            Some(caps_filter) => {
                let caps: gst::Caps;
                if name == "output" {
                    match gst::Caps::from_str(&*format!("video/x-h264,width={},height={},stream-format=byte-stream", width, height)) {
                        Ok(result) => {
                            caps = result;
                        }
                        Err(_) => {
                            return Err("FAILED TO BUILD OUTPUT CAPS");
                        }
                    }
                } else {
                    match gst::Caps::from_str(&*format!("video/x-raw,format=I420,width={},height={}", width, height)) {
                        Ok(result) => {
                            caps = result;
                        }
                        Err(_) => {
                            return Err("FAILED TO BUILD VIDEO CAPS");
                        }
                    }
                }
                caps_filter.set_property("caps", caps);
                Ok(())
            }
        };
    }

    pub fn change_video_position(&self, pad_name: &str, pos_x: i32, pos_y: i32) -> Result<(), &str> {
        return match self.compositor.static_pad(pad_name) {
            None => {
                Err("FAILED TO FIND COMPOSITOR STATIC PAD")
            }
            Some(sink_pad) => {
                sink_pad.set_property("xpos", pos_x);
                sink_pad.set_property("ypos", pos_y);
                sink_pad.set_property("zorder", 1000);

                sink_pad.set_property("width", 800);
                sink_pad.set_property("height", 800);
                Ok(())
            }
        };
    }

    pub fn focus_webcam(&self, pad_name: &str) -> Result<(), &str> {
        return self.change_video_position(pad_name, 0, 0);
        // self.change_video_size(Some(username), 800, 800);
    }

    // pub fn refresh_webcam(&self, sinks: &[SinkPad], tracks: &[&Track]) -> Vec<SinkPad> {
    //     let width: i32;
    //     let height: i32;
    //     if sinks.len() > 4 {
    //         width = 200;
    //         height = 200;
    //     } else {
    //         width = 400;
    //         height = 400;
    //     }
    //     let num = self.row_video_count(sinksunwrap.len());
    //     self.change_video_size(None, num * width, num * height);
    //
    //     for t in tracks {
    //         self.change_video_size(Some(t.username.as_str()), width, height);
    //     }
    //
    //     return self.relayout_positions(width, height);
    // }

    pub fn relayout_positions(&self, socket: Arc<Mutex<SplitSink<WebSocketStream<TcpStream>, Message>>>, width: i32, height: i32) -> Result<(), &str> {
        // self.pause().unwrap();
        let pads = self.compositor.sink_pads();
        // println!("pads.len {:?}", pads.len());
        if pads.is_empty() {
            return Err("COMPOSITOR SINK PADS IS EMPTY");
        }

        // let width: i32;
        // let height: i32;
        // if pads.len() > 4 {
        //     width = 200;
        //     height = 200;
        // } else {
        //     width = 400;
        //     height = 400;
        // }
        // let num = self.row_video_count(pads.len());
        // self.change_video_size(None, num * width, num * height);

        let mut sink_pads: Vec<SinkPad> = vec![];

        //todo each camera width => output / width
        //todo each camera height => output / height
        let num: i32 = self.row_video_count(pads.len());
        for (i, _) in pads.iter().enumerate() {
            //p("x :", i%2, "y :", i/2)
            pads[i].set_property("xpos", (i as i32 % num) * width);
            pads[i].set_property("ypos", (i as i32 / num) * height);
            pads[i].set_property_from_str("zorder", "0");
            // pads[i].set_property("width", w);
            // pads[i].set_property("height", h);

            sink_pads.push(SinkPad {
                name: pads[i].name().to_string(),
                x: ((i as i32 % num) * width) as i32,
                y: ((i as i32 / num) * height) as i32,
                // ..Default::default()
            })
        }
        // self.start().unwrap();

        let session_key = self.session_key.clone();
        futures::executor::block_on(async {
            let mut ws = socket.lock().await;
            return match serde_json::to_string(&sink_pads) {
                Err(_) => {
                    Err("FAILED TO SINK PADS TO SERIALIZE TO JSON")
                }
                Ok(sinks) => {
                    let msg = WsMessage {
                        event: models::EVENT_SINK_PADS.to_string(),
                        session: session_key,
                        data: sinks,
                    };
                    return match serde_json::to_string(&msg) {
                        Err(_) => {
                            Err("FAILED TO SINK PADS TO SERIALIZE TO JSON")
                        }
                        Ok(msg) => {
                            if let Err(e) = ws.send(Message::Text(msg)).await {
                                Err("FAILED TO SEND WS MESSAGE")
                            } else {
                                Ok(())
                            }
                        }
                    };
                }
            };
        })
    }

    pub async fn write_sample(&self, track: Arc<TrackLocalStaticSample>) -> Result<(), &str> {
        match self.pipeline.by_name("appsink") {
            None => {
                return Err("FAILED TO FIND APPSINK ELEMENT");
            }
            Some(appsink) => {
                let appsink_cast = appsink.dynamic_cast::<gst_app::AppSink>().unwrap();
                appsink_cast.set_state(gst::State::Playing).unwrap();

                loop {
                    match appsink_cast.pull_sample().map_err(|_| gst::FlowError::Eos) {
                        Err(_) => {
                            return Err("FAILED TO PULL SAMPLE FROM APPSINK");
                        }
                        Ok(sample) => {
                            let buffer = sample.buffer().unwrap().to_owned();
                            let map = buffer.map_readable().unwrap().to_owned();
                            let slice = map.as_slice().to_owned();
                            if let Some(duration) = buffer.duration() {
                                if let Err(e) = track.write_sample(&Sample {
                                    data: Bytes::from(slice),
                                    duration: Duration::from(duration),
                                    ..Sample::default()
                                }).await {
                                    println!("{:?}", e);
                                    return Err("FAILED TO WRITE SAMPLE TO TRACK");
                                };
                            }
                        }
                    };
                }

                // let cloned = appsink_cast.clone();
                // appsink_cast.set_callbacks(gst_app::AppSinkCallbacks::builder().new_sample(move |appsink| {
                //     let sample = cloned.to_owned().pull_sample().map_err(|_| gst::FlowError::Eos).unwrap().to_owned();
                //     let buffer = sample.buffer().unwrap().to_owned();
                //     let map = buffer.map_readable().unwrap().to_owned();
                //     let slice = map.as_slice().to_owned();
                //     futures::executor::block_on(async {
                //         track.write_sample(&Sample {
                //             data: Bytes::from(slice),
                //             duration: Duration::from(buffer.duration().unwrap()),
                //             ..Sample::default()
                //         }).await.unwrap();
                //     });
                //     Ok(gst::FlowSuccess::Ok)
                // }).build());
            }
        };
        // return Ok(());
    }

    pub fn send_eos_request(&self) {
        self.pipeline.send_event(gst::event::Eos::new());
    }

    pub async fn set_bus_listener(&self, pc: Weak<RTCPeerConnection>) -> Result<(), &str> {
        let bus = match self.pipeline.bus() {
            None => {
                return Err("FAILED TO STOP THE PIPELINE");
            }
            Some(v) => v
        };

        for msg in bus.iter_timed(gst::ClockTime::NONE) {
            use gst::MessageView;

            match msg.view() {
                MessageView::Eos(msg) => {
                    println!("EOS RECEIVED from {:?}", msg.src().unwrap().name());

                    let mut sessions = SESSIONS.lock().await;
                    let i = sessions.iter().position(|s| s.session_key == self.session_key);
                    if i.is_some() {
                        let i = i.unwrap();
                        let p = sessions.swap_remove(i);

                        for element in p.pipeline.iterate_elements() {
                            if let Ok(el) = element {
                                el.set_state(gst::State::Null).unwrap();
                                p.pipeline.remove(&el).unwrap();
                            }
                        }

                        p.stop().unwrap();
                    }

                    let connection = match pc.upgrade() {
                        None => {
                            println!("FAILED TO UPGRADE DOWNGRADED PEER CONNECTION");
                            break;
                        }
                        Some(v) => v
                    };
                    let senders = connection.get_senders().await;
                    if senders.len() == 0 { break; }
                    for sender in senders {
                        if let Some(track) = sender.track().await {
                            if track.id() == self.session_key {
                                if let Err(e) = connection.remove_track(&sender).await {
                                    println!("{}", e);
                                }
                                break;
                            }
                        }
                    }

                    break;
                }
                MessageView::Error(err) => {
                    println!("Error from {:?}: {} ({:?})", err.src().map(|s| s.path_string()), err.error(), err.debug());
                }
                // MessageView::StateChanged(state) => {
                //     println!("{:?} : State changed from {:?} to {:?}", state.src().unwrap().name(), state.old(), state.current());
                // }
                _ => (),
            }
        }

        return Ok(());
    }

    pub fn add_webcam(&self, username: &String) -> Result<(), &str> {
        let mut elements = vec![];

        if let Ok(appsrc) = gst::ElementFactory::make("appsrc")
            .name(&format!("{}_src", username))
            .property("is-live", true)
            .property("do-timestamp", true)
            .property_from_str("format", FORMAT_TIME) //time in nanoseconds
            .property("caps", gst::Caps::from_str(VIDEO_CAPS).unwrap())
            .build() {
            elements.push(appsrc);
        } else {
            return Err("FAILED TO MAKE APPSRC ELEMENT");
        }

        if let Ok(queue) = gst::ElementFactory::make("queue").property("silent", true)
            .name(&format!("{}_queue", username)).build() {
            elements.push(queue);
        } else {
            return Err("FAILED TO MAKE QUEUE ELEMENT");
        }

        // if let Ok(rtpjitterbuffer) = gst::ElementFactory::make("rtpjitterbuffer")
        //     //     .property_from_str("latency", "0")
        //     .name(&format!("{}_rtpjitterbuffer", username)).build() {
        //     elements.push(rtpjitterbuffer);
        // } else {
        //     return Err("FAILED TO MAKE RTPJITTERBUFFER ELEMENT");
        // }

        if let Ok(rtp_h264depay) = gst::ElementFactory::make("rtph264depay")
            .name(&format!("{}_depay", username)).build() {
            elements.push(rtp_h264depay);
        } else {
            return Err("FAILED TO MAKE RTP_H264DEPAY ELEMENT");
        }

        if let Ok(avdec_h264) = gst::ElementFactory::make("avdec_h264")
            .name(&format!("{}_decoder", username)).build() {
            elements.push(avdec_h264);
        } else {
            return Err("FAILED TO MAKE AVDEC_H264 ELEMENT");
        }

        if let Ok(videoscale) = gst::ElementFactory::make("videoscale")
            .name(&format!("{}_scaler", username)).build() {
            elements.push(videoscale);
        } else {
            return Err("FAILED TO MAKE VIDEOSCALE ELEMENT");
        }

        if let Ok(caps) = gst::Caps::from_str(&*format!("video/x-raw,format=I420,width={},height={},pixel-aspect-ratio=1/1", 300, 300)) {
            if let Ok(caps_filter) = gst::ElementFactory::make("capsfilter")
                .name(&*format!("{}_caps", username)).property("caps", caps).build() {
                elements.push(caps_filter);
            } else {
                return Err("FAILED TO MAKE CAPS FILTER ELEMENT");
            }
        } else {
            return Err("FAILED TO MAKE CAPS");
        }

        if let Ok(text_overlay) = gst::ElementFactory::make("textoverlay")
            .name(&format!("{}_text", username))
            .property_from_str("text", &username)
            .property_from_str("valignment", TEXT_VERTICAL_BOTTOM)
            .property_from_str("halignment", TEXT_HORIZONTAL_CENTER)
            .property_from_str("font-desc", "Arial, 46")
            .property_from_str("color", TEXT_WHITE_COLOR).build() {
            elements.push(text_overlay);
        } else {
            return Err("FAILED TO MAKE TEXT OVERLAY ELEMENT");
        }

        self.pipeline.add(elements.get(0).unwrap()).unwrap();
        for element in elements.windows(2) {
            self.pipeline.add(&element[1]).unwrap();
            element[0].link(&element[1]).unwrap();
        }

        if let Some(pad) = &self.compositor.request_pad_simple("sink_%u") {
            if let Err(_) = elements.get(elements.len() - 1).unwrap().static_pad("src").unwrap().link(pad) {
                return Err("FAILED TO LINK TEXT OVERLAY TO COMPOSITOR");
            }
        } else {
            return Err("FAILED TO GET NEW SINK PAD FROM COMPOSITOR");
        }

        for e in elements {
            e.set_state(gst::State::Playing).unwrap();
        }

        // self.pipeline.debug_to_dot_file(DebugGraphDetails::ALL, "test");

        Ok(())
    }
}
