use std::env;
use std::sync::Arc;

use dotenv::dotenv;
use futures_util::{SinkExt, StreamExt};
use futures_util::stream::{SplitSink, SplitStream};
use once_cell::sync::Lazy;
use tokio::net::{TcpListener, TcpStream};
use tokio::sync::Mutex;
use tokio_tungstenite::tungstenite::Message;
use tokio_tungstenite::WebSocketStream;
use webrtc::ice_transport::ice_candidate::RTCIceCandidateInit;
use webrtc::peer_connection::RTCPeerConnection;
use webrtc::peer_connection::sdp::session_description::RTCSessionDescription;

use models::WsMessage;

use crate::rtc::SESSIONS;

mod models;
mod gstreamer;
mod rtc;

pub static PEER_CONNECTION_MUTEX: Lazy<Arc<Mutex<Option<Arc<RTCPeerConnection>>>>> = Lazy::new(|| Arc::new(Mutex::new(None)));

async fn init_ws() -> Result<(SplitSink<WebSocketStream<TcpStream>, Message>, SplitStream<WebSocketStream<TcpStream>>), String> {
    let port = env::var("WS_PORT").unwrap_or("8080".to_string());
    if let Ok(listener) = TcpListener::bind(format!("0.0.0.0:{}", port)).await {
        if let Ok((stream, _)) = listener.accept().await {
            if let Ok(ws_stream) = tokio_tungstenite::accept_async(stream).await {
                return Ok(ws_stream.split());
            }
            return Err("FAILED TO ACCEPT ASYNC STREAM".to_string());
        }
        return Err("FAILED TO SET WS LISTENER".to_string());
    }
    return Err("FAILED TO CREATE WEBSOCKET SERVER".to_string());
}

#[tokio::main]
async fn main() -> std::io::Result<()> {
    dotenv().ok();
    println!("CITRA MEDIA SERVER STARTED ...");
    let (sender, mut receiver) = init_ws().await.unwrap();
    let pending_candidates: Arc<Mutex<Vec<RTCIceCandidateInit>>> = Arc::new(Mutex::new(vec![]));
    let socket = Arc::new(Mutex::new(sender));
    let peer_connection = rtc::init(&socket, &pending_candidates).await.unwrap();

    {
        let mut pcm = PEER_CONNECTION_MUTEX.lock().await;
        *pcm = Some(Arc::clone(&peer_connection));
    }

    gstreamer::init();
    loop {
        tokio::select! {
        msg = receiver.next() => {
            let text = match msg {
                Some(v) => v.unwrap().to_string(),
                None => {
                    println!("INVALID JSON MESSAGE");
                    continue;
                }
            };

            let message = match serde_json::from_str::<WsMessage>(&text) {
                Ok(v) => v,
                Err(_) => {
                    println!("INVALID JSON MESSAGE");
                    continue;
                }
            };
            match message.event.as_str() {
                models::EVENT_OFFER => {
                    let offer = match RTCSessionDescription::offer(message.data) {
                        Ok(v) => v,
                        Err(err) => {
                            println!("{:?}", err);
                            continue;
                        }
                    };
                    let pc = {
                        let pcm = PEER_CONNECTION_MUTEX.lock().await;
                        pcm.clone().unwrap()
                    };
                    if let Err(err) = pc.set_remote_description(offer).await {
                        println!("{:?}", err);
                        return Ok(());
                    }

                    let answer = match pc.create_answer(None).await {
                        Ok(a) => a,
                        Err(err) => {
                            println!("{:?}", err);
                            RTCSessionDescription::default()
                        }
                    };

                    if let Err(err) = pc.set_local_description(answer.clone()).await {
                        println!("{:?}", err);
                        return Ok(());
                    };
                    let answer_msg = WsMessage {
                        event: models::EVENT_ANSWER.to_string(),
                        data: answer.sdp,
                        ..Default::default()
                    };
                    if let Ok(ans) = serde_json::to_string(&answer_msg) {
                        if let Err(_) = socket.lock().await.send(Message::Text(ans)).await {
                            return Ok(());
                        }
                    } else {
                        return Ok(());
                    }

                    let mut cs = pending_candidates.lock().await;
                    for c in 0..cs.len() {
                        if let Err(err) = pc.add_ice_candidate(cs.pop().unwrap()).await {
                            println!("{:?}", err);
                            return Ok(());
                        }
                    }
                }
                models::EVENT_ANSWER => {
                    let pc = {
                        let pcm = PEER_CONNECTION_MUTEX.lock().await;
                        pcm.clone().unwrap()
                    };
                    if let Ok(answer) = RTCSessionDescription::answer(message.data) {
                        if let Err(err) = pc.set_remote_description(answer).await {
                            println!("{:?}", err);
                            return Ok(());
                        }
                    } else {
                        println!("FAILED TO PARSE ANSWER");
                    }
                }
                models::EVENT_CANDIDATE => {
                    let pc = {
                        let pcm = PEER_CONNECTION_MUTEX.lock().await;
                        pcm.clone().unwrap()
                    };
                    if let Ok(candidate) = serde_json::from_str::<RTCIceCandidateInit>(&message.data) {
                        let desc = pc.remote_description().await;
                        if desc.is_none() {
                            let mut cs = pending_candidates.lock().await;
                            cs.push(candidate);
                        } else {
                            if let Err(err) = pc.add_ice_candidate(candidate).await {
                                println!("{:?}", err);
                                return Ok(());
                            }
                        }
                    } else {
                        println!("FAILED TO PARSE CANDIDATE");
                    }
                }
                models::EVENT_FOCUS_WEBCAM => {
                    if let Ok(sink) = serde_json::from_str::<WsMessage>(&message.data) {
                        {
                            let sessions = SESSIONS.lock().await;
                            let s = sessions.iter().find(|s| s.session_key == sink.session);
                            if s.is_some() {
                                let s = s.unwrap();
                                if let Err(err) = s.focus_webcam(sink.data.as_str()) {
                                    println!("FAILED TO FOCUS WEBCAM : {}", err)
                                }
                            }
                        }
                    } else {
                        println!("FAILED TO PARSE FOCUS WEBCAM SINK");
                    }
                }
                models::EVENT_UNFOCUS_WEBCAM => {
                    if let Ok(sink) = serde_json::from_str::<WsMessage>(&message.data) {
                        {
                            let sessions = SESSIONS.lock().await;
                            let s = sessions.iter().find(|s| s.session_key == sink.session);
                            if s.is_some() {
                                let s = s.unwrap();
                                if let Err(e) = s.relayout_positions(socket.clone(), 300, 300) {
                                    println!("FAILED TO RELAYOUT POSITIONS : {:?}", e);
                                }
                            }
                        }
                    } else {
                        println!("FAILED TO PARSE UNFOCUS WEBCAM SINK");
                    }
                }
                _ => {
                    if let Err(e) = socket.lock().await.send(Message::Text("unknown msg".to_string())).await {
                        println!("FAILED TO SEND WS MESSAGE : {:?}", e);
                    }
                }
            }
        }
        }
    }

    // Ok(())
}
