use std::sync::{Arc, Weak};

use futures_util::stream::SplitSink;
use gst::Element;
use gst::glib::{Cast, ObjectExt};
use gst::prelude::{ElementExt, ElementExtManual, GObjectExtManualGst, GstBinExt, GstObjectExt, PadExt, PipelineExt};
use serde::{Deserialize, Serialize};
use tokio::net::TcpStream;
use tokio::sync::Mutex;
use tokio_tungstenite::tungstenite::Message;
use tokio_tungstenite::WebSocketStream;

#[macro_export]
macro_rules! upgrade_weak {
    ($x:ident, $r:expr) => {{
        match $x.upgrade() {
            Some(o) => o,
            None => return $r,
        }
    }};
    ($x:ident) => {
        upgrade_weak!($x, ())
    };
}

pub const EVENT_OFFER: &str = "9";
pub const EVENT_ANSWER: &str = "0";
pub const EVENT_CANDIDATE: &str = "3";
pub const EVENT_SINK_PADS: &str = "33";
pub const EVENT_FOCUS_WEBCAM: &str = "34";
pub const EVENT_UNFOCUS_WEBCAM: &str = "35";

const ULTRA_FAST: &str = "1";
const ZERO_LATENCY: &str = "0x00000004";
const COMPOSITOR_TRANSPARENT_BLACK: &str = "1";

#[derive(Default, Debug, Serialize, Deserialize)]
pub struct WsMessage {
    #[serde(default, rename = "event")]
    pub event: String,
    #[serde(default, rename = "session")]
    pub session: String,
    #[serde(default, rename = "data")]
    pub data: String,
}

#[derive(Default, Debug, Serialize, Deserialize)]
pub struct SinkPad {
    // #[serde(skip)]
    // pub username: String,

    // #[serde(skip)]
    // #[serde(skip_serializing)]
    // #[serde(skip_deserializing)]
    // #[serde(rename(serialize = "ser_name", deserialize = "de_name"))]
    #[serde(rename = "name")]
    pub name: String,

    #[serde(rename = "x")]
    pub x: i32,

    #[serde(rename = "y")]
    pub y: i32,
}

#[derive(Debug, Clone)]
pub struct Session(Arc<SessionInner>);

#[derive(Debug, Clone)]
pub struct SessionWeak(Weak<SessionInner>);

#[derive(Debug)]
pub struct SessionInner {
    pub session_key: String,
    pub pipeline: gst::Pipeline,
    pub compositor: gst::Element,
}

impl std::ops::Deref for Session {
    type Target = SessionInner;

    fn deref(&self) -> &SessionInner {
        &self.0
    }
}

impl SessionWeak {
    pub fn upgrade(&self) -> Option<Session> {
        self.0.upgrade().map(Session)
    }
}

impl Drop for SessionInner {
    fn drop(&mut self) {
        let _ = self.pipeline.set_state(gst::State::Null);
    }
}

impl Session {
    pub fn downgrade(&self) -> SessionWeak {
        SessionWeak(Arc::downgrade(&self.0))
    }

    pub fn new(socket: Arc<Mutex<SplitSink<WebSocketStream<TcpStream>, Message>>>, session_key: String) -> Result<Session, String> {
        let pipeline = gst::Pipeline::new(Some(session_key.as_str()));

        let (compositor, x264enc, appsink): (Element, Element, Element);
        let mut result = gst::ElementFactory::make("compositor").name("compositor")
            .property_from_str("background", COMPOSITOR_TRANSPARENT_BLACK)
            .build();

        if result.is_err() {
            return Err("FAILED TO MAKE COMPOSITOR ELEMENT".to_string());
        } else {
            compositor = result.unwrap();
        }

        pipeline.set_auto_flush_bus(true);
        pipeline.use_clock(Some(&gst::SystemClock::obtain()));
        pipeline.set_start_time(gst::ClockTime::NONE);
        pipeline.set_base_time(gst::ClockTime::ZERO);
        compositor.set_property_from_str("start-time-selection", "first");

        // openh264enc
        result = gst::ElementFactory::make("x264enc")
            .property_from_str("speed-preset", ULTRA_FAST)
            .property_from_str("tune", ZERO_LATENCY)
            .property_from_str("key-int-max", "20")
            .build();

        if result.is_err() {
            return Err("FAILED TO MAKE X264ENC ELEMENT".to_string());
        } else {
            x264enc = result.unwrap();
        }

        result = gst::ElementFactory::make("appsink")
            .property("sync", false)
            .name("appsink").build();

        if result.is_err() {
            return Err("FAILED TO MAKE APPSINK ELEMENT".to_string());
        } else {
            appsink = result.unwrap();
        }

        let session = Session(Arc::new(SessionInner {
            session_key,
            pipeline,
            compositor,
        }));

        let elements = &[&session.compositor, &x264enc, &appsink];

        session.pipeline.add(elements[0]).unwrap();
        for element in elements.windows(2) {
            session.pipeline.add(element[1]).unwrap();
            element[0].link(element[1]).unwrap();
        }

        let socket_link_clone = Arc::clone(&socket);
        let socket_unlink_clone = Arc::clone(&socket);

        let session_clone = session.downgrade();
        session.compositor.connect_pad_added(move |_bin, pad| {
            let session = upgrade_weak!(session_clone);

            let socket1 = socket_link_clone.clone();
            let socket2 = socket_unlink_clone.clone();

            let pipe_link_dg = session.downgrade();
            let pipe_unlink_dg = session.downgrade();

            pad.connect_linked(move |_, _| {
                let session = upgrade_weak!(pipe_link_dg);

                if let Err(err) = session.relayout_positions(socket1.clone(), 300, 300) {
                    // println!("FAILED TO RELAYOUT POSITIONS : {}", err);
                }
            });
            pad.connect_unlinked(move |_pad, _| {
                let session = upgrade_weak!(pipe_unlink_dg);

                if let Some(videomixer) = _pad.parent() {
                    let videomixer = videomixer.downcast_ref::<gst::Element>().unwrap();
                    videomixer.release_request_pad(_pad);
                }
                if let Err(err) = session.relayout_positions(socket2.clone(), 300, 300) {
                    // println!("FAILED TO RELAYOUT POSITIONS : {}", err);
                }
            });
        });

        for e in elements {
            e.set_state(gst::State::Playing).unwrap();
        }

        return Ok(session);
    }
}
