use std::env;
use std::io::Write;
use std::sync::Arc;
use std::time::Duration;

use anyhow::Result;
use bytes::BufMut;
use futures_util::SinkExt;
use futures_util::stream::SplitSink;
use gst::prelude::{Cast, ElementExt, ElementExtManual, GstBinExt, GstBinExtManual, GstObjectExt};
use once_cell::sync::Lazy;
use tokio::net::{TcpStream, UdpSocket};
use tokio::sync::Mutex;
use tokio_tungstenite::tungstenite::Message;
use tokio_tungstenite::WebSocketStream;
use webrtc::api::APIBuilder;
use webrtc::api::interceptor_registry::register_default_interceptors;
use webrtc::api::media_engine::{MediaEngine, MIME_TYPE_H264};
use webrtc::api::setting_engine::SettingEngine;
// use webrtc::Error as WEBRTCError;
use webrtc::ice::network_type::NetworkType;
use webrtc::ice::udp_mux::{UDPMuxDefault, UDPMuxParams};
use webrtc::ice::udp_network::UDPNetwork;
use webrtc::ice_transport::ice_candidate::{RTCIceCandidate, RTCIceCandidateInit};
use webrtc::ice_transport::ice_connection_state::RTCIceConnectionState;
use webrtc::interceptor::registry::Registry;
use webrtc::peer_connection::configuration::RTCConfiguration;
use webrtc::peer_connection::peer_connection_state::RTCPeerConnectionState;
use webrtc::peer_connection::policy::bundle_policy::RTCBundlePolicy;
use webrtc::peer_connection::policy::ice_transport_policy::RTCIceTransportPolicy;
use webrtc::peer_connection::policy::rtcp_mux_policy::RTCRtcpMuxPolicy;
use webrtc::peer_connection::RTCPeerConnection;
use webrtc::peer_connection::signaling_state::RTCSignalingState;
use webrtc::rtcp::payload_feedbacks::picture_loss_indication::PictureLossIndication;
use webrtc::rtp_transceiver::RTCRtpTransceiverInit;
use webrtc::rtp_transceiver::rtp_codec::{RTCRtpCodecCapability, RTCRtpCodecParameters, RTPCodecType};
use webrtc::rtp_transceiver::rtp_receiver::RTCRtpReceiver;
use webrtc::rtp_transceiver::rtp_transceiver_direction::RTCRtpTransceiverDirection;
use webrtc::track::track_local::track_local_static_sample::TrackLocalStaticSample;
use webrtc::track::track_local::TrackLocal;
use webrtc::track::track_remote::TrackRemote;

use crate::models::{EVENT_CANDIDATE, EVENT_OFFER, Session, WsMessage};
use crate::upgrade_weak;

pub static SESSIONS: Lazy<Mutex<Vec<Session>>> = Lazy::new(|| Mutex::new(vec![]));
// pub static SINK_PADS: Lazy<Mutex<Vec<Arc<SinkPad>>>> = Lazy::new(|| Mutex::new(vec![]));
// pub static PENDING_CANDIDATES: Lazy<Arc<Mutex<Vec<RTCIceCandidate>>>> = Lazy::new(|| Arc::new(Mutex::new(vec![])));

pub async fn init(socket: &Arc<Mutex<SplitSink<WebSocketStream<TcpStream>, Message>>>, _pending_candidates: &Arc<Mutex<Vec<RTCIceCandidateInit>>>) -> Result<Arc<RTCPeerConnection>, String> {
    let mut setting_engine = SettingEngine::default();
    let port = env::var("ICE_PORT").unwrap_or("8080".to_string()).parse::<u16>().unwrap_or(8080);
    let udp_socket = match UdpSocket::bind(("0.0.0.0", port)).await {
        Err(e) => return Err("FAILED TO BIND UDP SOCKET".to_string()),
        Ok(v) => v
    };

    let udp_mux = UDPMuxDefault::new(UDPMuxParams::new(udp_socket));
    let net = UDPNetwork::Muxed(udp_mux);

    // let udp_ephemeral = EphemeralUDP::new(8080, 8080).unwrap();
    // let net = UDPNetwork::Ephemeral(udp_ephemeral);

    setting_engine.set_network_types(vec![NetworkType::Udp4]);
    setting_engine.set_udp_network(net);
    // setting_engine.disable_media_engine_copy(true);

    let mut media_engine = MediaEngine::default();
    // media_engine.register_default_codecs().unwrap();
    media_engine.register_codec(RTCRtpCodecParameters {
        capability: RTCRtpCodecCapability {
            mime_type: MIME_TYPE_H264.to_owned(),
            clock_rate: 90000,
            channels: 0,
            sdp_fmtp_line: "".to_owned(),
            rtcp_feedback: vec![],
        },
        payload_type: 96,
        ..Default::default()
    }, RTPCodecType::Video).unwrap();

    let mut registry = Registry::new();
    registry = register_default_interceptors(registry, &mut media_engine).unwrap();

    // if env::var("ENABLE_NACK").unwrap() == "true" {
    //     let generator = Box::from(nack::generator::GeneratorBuilder::default());
    //     let responder = Box::from(nack::responder::ResponderBuilder::default());
    //     registry.add(generator);
    //     registry.add(responder);
    // }

    let api = APIBuilder::new()
        .with_setting_engine(setting_engine)
        .with_media_engine(media_engine)
        .with_interceptor_registry(registry)
        .build();

    let config = RTCConfiguration {
        ice_servers: vec![],
        ice_transport_policy: RTCIceTransportPolicy::All,
        bundle_policy: RTCBundlePolicy::MaxBundle,
        rtcp_mux_policy: RTCRtcpMuxPolicy::Require,
        ..RTCConfiguration::default()
    };

    let peer_connection = Arc::new(api.new_peer_connection(config).await.unwrap());

    peer_connection.on_peer_connection_state_change(Box::new(move |s: RTCPeerConnectionState| {
        println!("Peer Connection State has changed: {}", s);
        // if s == RTCPeerConnectionState::Failed {
        //     println!("Peer Connection has gone to failed exiting");
        // }
        Box::pin(async {})
    }));

    peer_connection.on_signaling_state_change(Box::new(move |s: RTCSignalingState| {
        println!("Signaling State has changed: {}", s);
        // if s == RTCPeerConnectionState::Failed {
        //     println!("Peer Connection has gone to failed exiting");
        // }
        Box::pin(async {})
    }));

    peer_connection.on_ice_connection_state_change(Box::new(move |s: RTCIceConnectionState| {
        println!("ICE State has changed: {}", s);
        // if s == RTCPeerConnectionState::Failed {
        //     println!("Peer Connection has gone to failed exiting");
        // }
        Box::pin(async {})
    }));

    let ws1 = Arc::downgrade(&socket);
    let pc = Arc::downgrade(&peer_connection);
    peer_connection.on_negotiation_needed(Box::new(move || {
        let pc2 = pc.clone();
        let ws2 = ws1.clone();
        Box::pin(async move {
            if let Some(pc) = pc2.upgrade() {
                // Some(RTCOfferOptions { ice_restart: true, voice_activity_detection: false})
                if let Ok(offer) = pc.create_offer(None).await {
                    if let Err(e) = pc.set_local_description(offer.clone()).await {
                        println!("FAILED TO SET LOCAL DESCRIPTION : {:?}", e);
                        return;
                    }
                    let message = WsMessage {
                        event: EVENT_OFFER.to_string(),
                        data: offer.sdp,
                        ..Default::default()
                    };
                    if let Ok(sdp) = serde_json::to_string(&message) {
                        if let Some(upgraded_ws) = ws2.upgrade() {
                            let mut ws = upgraded_ws.lock().await;
                            if let Err(e) = ws.send(Message::Text(sdp)).await {
                                println!("FAILED TO SEND WS MESSAGE : {:?}", e);
                                return;
                            }
                        } else {
                            println!("FAILED TO UPGRADE DOWNGRADED WS");
                            return;
                        }
                    } else {
                        println!("FAILED TO SERIALIZE DATA TO JSON");
                        return;
                    }
                }
            }
        })
    }));
    let ws2 = Arc::downgrade(&socket);
    let pc = Arc::downgrade(&peer_connection);
    // let candidates = Arc::downgrade(&pending_candidates);
    peer_connection.on_ice_candidate(Box::new(move |c: Option<RTCIceCandidate>| {
        let ws3 = ws2.clone();
        let pc2 = pc.clone();
        // let candidates2 = candidates.clone();
        Box::pin(async move {
            if let Some(c) = c {
                if let Some(pc) = pc2.upgrade() {
                    // let desc = pc.remote_description().await;
                    // if desc.is_none() {
                    //     if let Some(candidate) = candidates2.upgrade() {
                    //         let mut cs = candidate.lock().await;
                    //         cs.push(c);
                    //     }
                    // } else {
                    if let Ok(ice_candidate) = c.to_json() {
                        if let Ok(data) = serde_json::to_string(&ice_candidate) {
                            let message = WsMessage {
                                event: EVENT_CANDIDATE.to_string(),
                                data,
                                ..Default::default()
                            };
                            if let Ok(candidate) = serde_json::to_string(&message) {
                                if let Some(upgraded_ws) = ws3.upgrade() {
                                    let mut ws = upgraded_ws.lock().await;
                                    match ws.send(Message::Text(candidate)).await {
                                        Ok(_) => {}
                                        Err(err) => {
                                            println!("FAILED TO SEND WS MESSAGE : {:?}", err);
                                        }
                                    };
                                } else {
                                    println!("FAILED TO SERIALIZE DATA TO JSON");
                                }
                            } else {
                                println!("FAILED TO UPGRADE DOWNGRADED WS");
                            }
                        }
                    } else {
                        println!("FAILED TO CREATE JSON FROM ICE CANDIDATE");
                    }
                    // }
                }
            }
        })
    }));

    let pc = Arc::downgrade(&peer_connection);
    let ws3 = Arc::downgrade(&socket);
    peer_connection.on_track(Box::new(move |remote_track: Arc<TrackRemote>, _receiver: Arc<RTCRtpReceiver>, _| {
        // println!("track received");

        let ws4 = ws3.clone();
        let remote_clone = Arc::clone(&remote_track);

        let media_ssrc = remote_track.ssrc();
        let pc2 = pc.clone();
        tokio::spawn(async move {
            let mut result = Result::<usize>::Ok(0);
            if let Some(pc) = pc2.upgrade() {
                while result.is_ok() {
                    let timeout = tokio::time::sleep(Duration::from_secs(1));
                    tokio::pin!(timeout);

                    tokio::select! {
                        _ = timeout.as_mut() =>{
                                result = pc.write_rtcp(&[Box::new(PictureLossIndication{
                                        sender_ssrc: 0,
                                        media_ssrc,
                                })]).await.map_err(Into::into);
                        }
                    }
                }
            }
        });

        let mut username = String::new();
        let mut session_key = String::new();
        let pc2 = pc.clone();

        Box::pin(async move {
            let track_id = remote_track.id();
            let parts: Vec<_> = track_id.split('|').collect();
            username = parts[0].to_string();
            session_key = parts[1].to_string();

            // println!("{} : {}", chrono::offset::Local::now().format("%Y-%m-%d %T"), username);

            let mut current_session = {
                let lock = SESSIONS.lock().await;
                let s = lock.iter().find(|s| s.session_key == session_key);
                if s.is_none() {
                    None
                } else {
                    let s = s.unwrap();
                    Some(s.clone())
                }
            };

            if current_session.is_none() {
                let track = Arc::new(TrackLocalStaticSample::new(
                    RTCRtpCodecCapability {
                        mime_type: MIME_TYPE_H264.to_owned(),
                        ..Default::default()
                    }, session_key.clone(), "camera".to_string(),
                ));

                let socket = match ws4.upgrade() {
                    None => {
                        println!("FAILED TO UPGRADE DOWNGRADED WS");
                        return;
                    }
                    Some(v) => v
                };
                let session = match Session::new(socket.clone(), session_key.clone()) {
                    Ok(v) => v,
                    Err(_) => {
                        println!("FAILED TO CREATE NEW SESSION");
                        return;
                    }
                };
                if let Err(e) = session.start() {
                    println!("FAILED TO START PIPELINE");
                    return;
                }
                let write_clone = session.downgrade();
                let bus_clone = session.downgrade();

                current_session = Some(session.clone());
                {
                    let mut vec = SESSIONS.lock().await;
                    vec.push(session);
                };

                let connection = match pc2.upgrade() {
                    None => {
                        println!("FAILED TO UPGRADE DOWNGRADED PEER CONNECTION");
                        return;
                    }
                    Some(v) => v
                };

                let transceiver = match connection.add_transceiver_from_track(track.clone() as Arc<dyn TrackLocal + Send + Sync>, Some(RTCRtpTransceiverInit {
                    direction: RTCRtpTransceiverDirection::Sendonly,
                    send_encodings: vec![],
                })).await {
                    Ok(v) => v,
                    Err(_) => {
                        println!("FAILED TO ADD TRANSCEIVER TO PC");
                        return;
                    }
                };

                tokio::spawn(async move {
                    let mut rtcp_buf = vec![0u8; 1500];
                    while let Ok((_, _)) = transceiver.sender().read(&mut rtcp_buf).await {}
                    Result::<()>::Ok(())
                });

                tokio::spawn(async move {
                    let session = upgrade_weak!(write_clone);
                    if let Err(e) = session.write_sample(track).await {
                        println!("{}", e);
                    }
                });

                tokio::spawn(async move {
                    let session = upgrade_weak!(bus_clone);
                    if let Err(e) = session.set_bus_listener(pc2).await {
                        println!("{}", e);
                    }
                });
            }

            let current_session = current_session.unwrap();
            if let Err(e) = current_session.add_webcam(&username) {
                println!("{}", e);
                return;
            }

            tokio::spawn(async move {
                let appsrc = match current_session.pipeline.by_name(&format!("{}_src", username)) {
                    None => {
                        println!("FAILED TO FIND APPSRC ELEMENT");
                        return;
                    }
                    Some(v) => v
                };

                if let Ok(appsrc) = appsrc.dynamic_cast::<gst_app::AppSrc>() {
                    let mut b: [u8; 1500] = [0; 1500];
                    while let Ok((n, _)) = remote_clone.read(&mut b).await {
                        let mut buffer = gst::Buffer::with_size(n).unwrap();
                        {
                            let buffer = buffer.get_mut().unwrap();
                            let mut writer_map = buffer.map_writable().unwrap();
                            let mut writer = writer_map.writer();
                            writer.write(&b[..n]).unwrap();
                        }
                        if let Err(e) = appsrc.push_buffer(buffer) {
                            println!("FAILED TO PUSH BUFFER TO APPSRC");
                            break;
                        }
                    }
                }

                let sessions = SESSIONS.lock().await;
                let session = sessions.iter().find(|s| s.session_key == session_key);
                if session.is_some() {
                    let session = session.unwrap();
                    if session.compositor.sink_pads().len() > 1 {
                        if let Err(e) = session.remove_webcam(username.as_str()) {
                            println!("{}", e);
                        }
                    } else {
                        session.send_eos_request();
                    }
                }
            });
        })
    }));
    Ok(peer_connection)
}
